from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import JobFuture, GatherJobsFuture, retry_fn
from ai2_kit.core.future import map_future
from ai2_kit.core.log import get_logger

from pydantic import BaseModel
from typing import List
from dataclasses import dataclass
import os
import copy
import random
import sys
import json

from .cll import ICllTrainInput, ICllTrainOutput, BaseCllContext
from .data_helper import Cp2kOutputHelper, DeepmdNpyHelper, DeepmdModelHelper, convert_to_deepmd_npy
from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
)

logger = get_logger(__name__)


class GenericDeepmdInputConfig(BaseModel):
    model_num: int = 4
    init_dataset: List[str]
    input_template: dict


class GenericDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'


@dataclass
class GenericDeepmdInput(ICllTrainInput):
    config: GenericDeepmdInputConfig
    type_map: List[str]
    old_dataset: List[Artifact]  # training data used by previous iteration
    new_dataset: List[Artifact]  # training data used by current iteration
    initiated: bool = False

    def update_training_dataset(self, dataset: List[Artifact]):
        self.old_dataset += self.new_dataset
        self.new_dataset = dataset

@dataclass
class GenericDeepmdContext(BaseCllContext):
    config: GenericDeepmdContextConfig

@dataclass
class GenericDeepmdOutput(ICllTrainOutput):
    outputs: List[Artifact]
    input: GenericDeepmdInput

    def get_mlp_models(self) -> List[Artifact]:
        return [a.join(DP_FROZEN_MODEL) for a in self.outputs]

    def get_training_dataset(self) -> List[Artifact]:
        return self.input.new_dataset + self.input.old_dataset

def generic_deepmd(input: GenericDeepmdInput, ctx: GenericDeepmdContext):
    executor = ctx.resource_manager.default_executor

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [converted_data_dir, tasks_dir] = executor.setup_workspace(work_dir, ['converted_input_data', 'tasks'])

    # initialization
    if not input.initiated:
        input.new_dataset += ctx.resource_manager.resolve_artifacts(input.config.init_dataset)

    # convert data type if necessary, only needed for new data as old data is already converted
    new_deepmd_npy_data: List[Artifact] = []
    cp2k_output_data: List[Artifact] = []

    # TODO: refactor data type conversion
    for artifact in input.new_dataset:
        if not artifact.format or DeepmdNpyHelper.is_match(artifact):
            # treated as deepmd npy data if format is not specified
            new_deepmd_npy_data.append(artifact)
        elif Cp2kOutputHelper.is_match(artifact):
            cp2k_output_data.append(artifact)
        else:
            raise ValueError(f'unsupported data type: {artifact.format}')

    # convert data to deepmd/npy format
    # TODO: support more data type
    converted_data_dirs = executor.run_python_fn(convert_to_deepmd_npy)(
        cp2k_outputs=[a.to_dict() for a in cp2k_output_data],
        base_dir=converted_data_dir,
        type_map=input.type_map,
    )
    new_deepmd_npy_data += [Artifact.of(
        url=url, format=DeepmdNpyHelper.format,
    ) for url in converted_data_dirs]

    input.new_dataset = new_deepmd_npy_data

    jobs: List[JobFuture] = []
    output_dirs = []

    # train multiple models at the same time
    for i in range(input.config.model_num):
        # each model should be trained in its own task_dir
        task_dir = os.path.join(tasks_dir, str(i).zfill(3))
        executor.mkdir(task_dir)

        # create dp train input file
        # NOTE: dp v1 format is supported currently
        # TODO: support more params if it is necessary

        # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

        dp_input = copy.deepcopy(input.config.input_template)
        training: dict = dp_input['training']

        # set output files
        training['disp_file'] = DP_DISP_FILE
        training['save_ckpt'] = DP_CHECKPOINT_FILE
        training['profiling_file'] = DP_PROFILING_FILE

        # set random seed
        discriptor = dp_input['model']['descriptor']
        if discriptor['type'] == 'hybrid':
            for d in discriptor['list']:
                d['seed'] = _random_seed()
        else:
            discriptor['seed'] = _random_seed()
        dp_input['model']['fitting_net']['seed'] = _random_seed()
        dp_input['training']['seed'] = _random_seed()

        # set training data
        systems = [a.url for a in input.old_dataset + input.new_dataset]
        training['systems'] = systems
        set_prefix: str = training.setdefault(
            'set_prefix', 'set')  # respect user input
        auto_prob_str = "prob_sys_size"
        training.setdefault('batch_size', 'auto')
        training['auto_prob_style'] = auto_prob_str

        # v2 training data
        training_data = {
            'systems': training['systems'],
            'set_prefix': training['set_prefix'],
            'auto_prob_style': training['auto_prob_style'],
            'batch_size': training['batch_size'],
        }
        training['training_data'] = training_data

        # other params
        dp_input['model']['type_map'] = input.type_map

        # write config to executor
        dp_input_text = json.dumps(dp_input, indent=2)
        dp_input_path = os.path.join(task_dir, DP_INPUT_FILE)
        executor.dump_text(dp_input_text, dp_input_path)

        # build script
        dp_cmd = ctx.config.dp_cmd
        dp_train_cmd = [dp_cmd, 'train', DP_INPUT_FILE]
        dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_FROZEN_MODEL]

        dp_train_script = BashScript(
            template=ctx.config.script_template,
            steps=[
                BashStep(cmd=dp_train_cmd, checkpoint='dp-train'),  # type: ignore
                BashStep(cmd=dp_freeze_cmd),  # type: ignore
            ] # type: ignore
        )
        output_dirs.append(task_dir)

        # submit job
        job = executor.submit(dp_train_script.render(), cwd=task_dir)
        jobs.append(job)

    future = GatherJobsFuture(jobs, done_fn=retry_fn(max_tries=2), raise_exception=True)

    return map_future(future, GenericDeepmdOutput(
        input=input,
        outputs=[Artifact.of(
            url=url,
        ) for url in output_dirs]
    ))

def _random_seed():
    return random.randrange(sys.maxsize) % (1 << 32)



from abc import abstractmethod, ABC
from ai2_kit.core.artifact import Artifact
from ai2_kit.core.future import IFuture
from ai2_kit.core.resource_manager import ResourceManager
from typing import List, Callable, Optional
from dataclasses import dataclass

@dataclass
class BaseCllContext:
    path_prefix: str
    resource_manager: ResourceManager

######################
# CLL Labeling Tasks #
######################

class ICllLabelInput(ABC):

    @abstractmethod
    def set_systems(self, systems: List[Artifact]):
        ...


class ICllLabelOutput(ABC):

    @abstractmethod
    def get_labeled_system_dataset(self) -> List[Artifact]:
        ...

CllLabelTaskType = Callable[[ICllLabelInput, BaseCllContext], IFuture[ICllLabelOutput]]

######################
# CLL Training Tasks #
######################

class ICllTrainInput(ABC):

    @abstractmethod
    def update_training_dataset(self, data: List[Artifact]):
        ...


class ICllTrainOutput(ABC):

    @abstractmethod
    def get_mlp_models(self) -> List[Artifact]:
        ...

    @abstractmethod
    def get_training_dataset(self) -> List[Artifact]:
        ...


CllTrainTaskType = Callable[[ICllTrainInput, BaseCllContext], IFuture[ICllTrainOutput]]

######################
# CLL Explore Tasks #
######################

class ICllExploreInput(ABC):

    @abstractmethod
    def set_md_models(self, models: List[Artifact]):
        ...

    @abstractmethod
    def set_fep_models(self, red_models: List[Artifact], neu_models: List[Artifact]):
        ...


class ICllExploreOutput(ABC):

    @abstractmethod
    def get_model_devi_dataset(self) -> List[Artifact]:
        ...


CllExploreTaskType = Callable[[ICllExploreInput, BaseCllContext], IFuture[ICllExploreOutput]]

######################
# CLL Select Task    #
######################

class ICllSelectorInput(ABC):

    @abstractmethod
    def set_model_devi_dataset(self, data: List[Artifact]):
        ...

class ICllSelectorOutput(ABC):

    @abstractmethod
    def get_model_devi_dataset(self) -> List[Artifact]:
        ...

    @abstractmethod
    def get_passing_rate(self) -> float:
        ...

CllSelectorTaskType = Callable[[ICllSelectorInput, BaseCllContext], IFuture[ICllSelectorOutput]]




from ruamel.yaml import YAML
from pathlib import Path
from typing import Tuple, List, TypeVar
from dataclasses import field
from cp2k_input_tools.parser import CP2KInputParserSimplified

import shortuuid
import copy
import os
import io

from .log import get_logger

logger = get_logger(__name__)

def default_mutable_field(obj):
    return field(default_factory=lambda: copy.copy(obj))

def __merge_dict():
    """cloudpickle compatible: https://stackoverflow.com/questions/75292769"""
    def merge_dict(lo: dict, ro: dict, path=None):
        """
        merge two dict, the left dict will be overridden

        this method won't merge list
        """
        if path is None:
            path = []

        for key, value in ro.items():
            if key in lo:
                current_path = path + [ str(key) ]
                if isinstance(lo[key], dict) and isinstance(value, dict):
                    merge_dict(lo[key], value, current_path)
                else:
                    print('.'.join(current_path) + ' has been overridden')
                    lo[key] = value
            else:
                lo[key] = value
        return lo
    return merge_dict
merge_dict = __merge_dict()


# TODO: support http(s) url
def load_yaml_file(path: Path):
    yaml = YAML(typ='safe')
    return yaml.load(path)


def load_yaml_files(*paths: Tuple[Path]):
    d = {}
    for path in paths:
        print('load yaml file: ', path)
        d = merge_dict(d, load_yaml_file(Path(path)))  # type: ignore
    return d


def s_uuid():
    """short uuid"""
    return shortuuid.uuid()


def parse_cp2k_input(text: str):
    parser = CP2KInputParserSimplified(key_trafo=str.upper)
    return parser.parse(io.StringIO(text))

def dict_nested_get(d: dict, keys: List[str]):
    """get value from nested dict"""
    for key in keys:
        d = d[key]
    return d

def dict_nested_set(d: dict, keys: List[str], value):
    """set value to nested dict"""
    for key in keys[:-1]:
        d = d[key]
    d[keys[-1]] = value

def sort_unique_str_list(l: List[str]) -> List[str]:
    """remove duplicate str and sort"""
    return list(sorted(set(l)))


T = TypeVar('T')
def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]


import ase.io
from ase import Atoms
from typing import List

class AseHelper:
    def __init__(self):
        self.atoms_list: List[Atoms] = []

    def read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self.atoms_list = data
        return self

    def write(self, filename: str, **kwargs):
        ase.io.write(filename, self.atoms_list, **kwargs)
        return self

    def write_each_frame(self, filename: str, **kwargs):
        for i, atoms in enumerate(self.atoms_list):
            ase.io.write(filename.format(i=i), atoms, **kwargs)
        return self

    def done(self):
        ...


from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import GatherJobsFuture, retry_fn

from ai2_kit.core.future import DummyFuture, IFuture, map_future

from ai2_kit.core.util import merge_dict, parse_cp2k_input, dict_nested_get, dict_nested_set
from ai2_kit.core.log import get_logger

from typing import List, Union
from pydantic import BaseModel
from dataclasses import dataclass

import copy
import os

from .data_helper import LammpsOutputHelper, XyzHelper, Cp2kOutputHelper, ase_atoms_to_cp2k_input_data
from .cll import ICllLabelInput, ICllLabelOutput, BaseCllContext

logger = get_logger(__name__)

class GenericCp2kInputConfig(BaseModel):
    init_system_files: List[str] = []
    limit: int = 50
    input_template: Union[dict, str]
    """
    Input template for cp2k. Could be a dict or content of a cp2k input file.

    Note:
    If you are using files in input templates, it is recommended to use artifact name instead of literal path.
    String starts with '@' will be treated as artifact name.
    For examples, FORCE_EVAL/DFT/BASIS_SET_FILE_NAME = @cp2k/basic_set.
    You can still use literal path, but it is not recommended.
    """

class GenericCp2kContextConfig(BaseModel):
    script_template: BashTemplate
    cp2k_cmd: str = 'cp2k'
    concurrency: int = 5

@dataclass
class GenericCp2kInput(ICllLabelInput):

    config: GenericCp2kInputConfig
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False

    def set_systems(self, systems: List[Artifact]):
        self.system_files = systems


@dataclass
class GenericCp2kContext(BaseCllContext):
    config: GenericCp2kContextConfig


@dataclass
class GenericCp2kOutput(ICllLabelOutput):
    cp2k_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.cp2k_outputs


def generic_cp2k(input: GenericCp2kInput, ctx: GenericCp2kContext) -> IFuture[GenericCp2kOutput]:

    executor = ctx.resource_manager.default_executor

    # For the first round
    if not input.initiated:
        input.system_files += ctx.resource_manager.get_artifacts(input.config.init_system_files)

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # prepare input template
    if isinstance(input.config.input_template, str):
        input_template = parse_cp2k_input(input.config.input_template)
    else:
        input_template = copy.deepcopy(input.config.input_template)

    fields_with_artifact = [
        ['FORCE_EVAL', 'DFT', 'BASIS_SET_FILE_NAME'],
        ['FORCE_EVAL', 'DFT', 'POTENTIAL_FILE_NAME'],
        ['FORCE_EVAL', 'DFT', 'XC', 'VDW_POTENTIAL', 'PAIR_POTENTIAL', 'PARAMETER', 'PARAMETER_FILE_NAME' ],
    ]
    for field_path in fields_with_artifact:
        try :
            field_value = dict_nested_get(input_template, field_path)
            if isinstance(field_value, str) and field_value.startswith('@'):
                logger.info(f'resolve artifact {field_value}')
                dict_nested_set(
                    input_template,
                    field_path,
                    ctx.resource_manager.resolve_artifact(field_value[1:])[0].url
                )
        except KeyError:
            pass

    # resolve data files
    lammps_dump_files: List[Artifact] = []
    xyz_files: List[Artifact] = []

    # TODO: support POSCAR in the future
    # TODO: refactor the way of handling different file formats
    for system_file in input.system_files:
        if LammpsOutputHelper.is_match(system_file):
            lammps_out = LammpsOutputHelper(system_file)
            lammps_dump_files.extend(lammps_out.get_passed_dump_files())
        elif XyzHelper.is_match(system_file):
            xyz_files.append(system_file)
        else:
            raise ValueError(f'unsupported format {system_file.url}: {system_file.format}')

    # create task dirs and prepare input files
    cp2k_task_dirs = []
    if lammps_dump_files or xyz_files:
        cp2k_task_dirs = executor.run_python_fn(make_cp2k_task_dirs)(
            lammps_dump_files=[a.url for a in lammps_dump_files],
            xyz_files=[a.url for a in xyz_files],
            type_map=input.type_map,
            base_dir=tasks_dir,
            input_template=input_template,
            limit=input.config.limit,
        )
    else:
        logger.warn('no available candidates, skip')
        return DummyFuture(GenericCp2kOutput(cp2k_outputs=[]))

    # group cp2k tasks by concurrency
    concurrency = ctx.config.concurrency
    steps_group = [list() for _ in range(concurrency)]
    for i, cp2k_task_dir in enumerate(cp2k_task_dirs):
        steps = steps_group[i % concurrency]
        step = BashStep(
            cwd=cp2k_task_dir,
            cmd=[ctx.config.cp2k_cmd, '-i input.inp 1>> output 2>> output'],
            checkpoint='cp2k',
        )
        steps.append(step)

    # run tasks
    jobs = []
    for steps in steps_group:
        if not steps:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps,
        )
        job = executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)

    future = GatherJobsFuture(jobs, done_fn=retry_fn(max_tries=2), raise_exception=True)

    cp2k_outputs = [Artifact.of(
        url=url,
        format=Cp2kOutputHelper.format,
        executor=executor.name,
        attrs=dict(),  # TODO: success from input
    ) for url in cp2k_task_dirs]

    return map_future(future, GenericCp2kOutput(cp2k_outputs=cp2k_outputs))


def __make_cp2k_task_dirs():
    def make_cp2k_task_dirs(lammps_dump_files: List[str],
                            xyz_files: List[str],
                            type_map: List[str],
                            input_template: dict,
                            base_dir: str,
                            limit: int = 0,
                            input_file_name: str = 'input.inp',
                            ) -> List[str]:
        """Generate CP2K input files from LAMMPS dump files or XYZ files."""

        # TODO: pymatgen has a better support for cp2k input, replace cp2k_input_tools with pymatgen in the future
        # https://pymatgen.org/pymatgen.io.cp2k.inputs.html#pymatgen.io.cp2k.inputs.Cp2kInput
        from cp2k_input_tools import DEFAULT_CP2K_INPUT_XML
        from cp2k_input_tools.generator import CP2KInputGenerator

        import ase.io
        from ase import Atoms

        cp2k_generator = CP2KInputGenerator(DEFAULT_CP2K_INPUT_XML)
        task_dirs = []
        atoms_list: List[Atoms] = []

        # read atoms
        for dump_file in lammps_dump_files:
            atoms_list += ase.io.read(dump_file, ':', format='lammps-dump-text', order=False, specorder=type_map)  # type: ignore
        for xyz_file in xyz_files:
            atoms_list += ase.io.read(xyz_file, ':', format='extxyz', order=False)  # type: ignore

        if limit > 0:
            atoms_list = atoms_list[:limit]

        for i, atoms in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)

            # create input file
            input_data = copy.deepcopy(input_template)
            coords, cell = ase_atoms_to_cp2k_input_data(atoms)
            merge_dict(input_data, {
                'FORCE_EVAL': {
                    'SUBSYS': {
                        'COORD': {
                            '*': coords
                        },
                        'CELL': {
                            'A': cell[0],
                            'B': cell[1],
                            'C': cell[2],
                        }
                    }
                }
            })
            input_text = '\n'.join(cp2k_generator.line_iter(input_data))
            with open(os.path.join(task_dir, input_file_name), 'w') as f:
                f.write(input_text)
            task_dirs.append(task_dir)
        return task_dirs

    return make_cp2k_task_dirs
make_cp2k_task_dirs = __make_cp2k_task_dirs()



from typing import Generic, Optional, TypeVar, Callable
from abc import abstractmethod


T = TypeVar('T')


# TODO: use RxPY Observable instead
# mimic the API of concurrent.future.Future
class IFuture(Generic[T]):

    @abstractmethod
    def done(self) -> bool:
        pass

    @abstractmethod
    def result(self, timeout: Optional[float] = None) -> T:
        pass

class DummyFuture(IFuture[T]):

    def __init__(self, value: T) -> None:
        self.value = value

    def done(self):
        return True

    def result(self, timeout: Optional[float] = None) -> T:
        return self.value


NT = TypeVar('NT')

MapFn = Callable[[T], NT]

class MapFuture(IFuture[NT]):

    def __init__(self, future: IFuture[T], map_fn: MapFn):
        self._future = future
        self._map_fn = map_fn

    def done(self):
        return self._future.done()

    def result(self, timeout=None) -> NT:
        result = self._future.result(timeout)
        return self._map_fn(result)

def map_future(future: IFuture, value: T):
    return MapFuture[T](future, lambda _ : value)


from typing import TypeVar, Union, Callable, NamedTuple, Optional
from threading import Lock
import functools
import cloudpickle
import os
import types

from .future import IFuture

_lock = Lock()
_checkpoint_file: Optional[str] = None
_checkpoint_data: Optional[dict] = None


class FnInfo(NamedTuple):
    fn_name: str
    args: tuple
    kwargs: dict


KeyFn = Callable[[FnInfo], str]

EMPTY = object()


def set_checkpoint_file(path: str):
    global _checkpoint_file
    if _checkpoint_file is not None:
        raise RuntimeError(
            "checkpoint path has been set to {}".format(_checkpoint_file))
    _checkpoint_file = path


def checkpoint(key_fn: Union[str, KeyFn], disable = False):
    """
    Checkpoint for function.

    Note: This checkpoint implementation doesn't support multiprocess.
    To support multiple process we need to have a dedicated background process to read/write checkpoint,
    which will require message queue (e.g. nanomsg or nng) to implement it.

    Example:

    >>> set_checkpoint_file('/tmp/test.ckpt')
    >>> task_fn = lambda a, b: a + b
    >>> checkpoint('task_1+2')(task_fn)(1, 2)
    """

    T = TypeVar('T', bound=Callable)
    def _checkpoint(fn: T) -> T:

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            if isinstance(key_fn, str):
                key = key_fn
            else:
                fn_info = FnInfo(
                    fn_name=fn.__name__,
                    args=args,
                    kwargs=kwargs,
                )
                key = key_fn(fn_info)

            if disable or _checkpoint_file is None:
                return fn(*args, **kwargs)

            ret = _get_checkpoint(key)
            if ret is not EMPTY:
                return ret

            ret = fn(*args, **kwargs)

            if isinstance(ret, IFuture):
                result_fn = ret.result
                def _wrap_fn(self, timeout=None):
                    _ret = result_fn(timeout)
                    _set_checkpoint(key, ret)
                    return _ret
                ret.result = types.MethodType(_wrap_fn, ret)
            else:
                _set_checkpoint(key, ret)

            return ret

        return wrapper # type: ignore

    return _checkpoint


def _load_checkpoint():
    global _checkpoint_data
    if _checkpoint_data is not None:
        return
    assert _checkpoint_file is not None, '_checkpoint_path should not be None!'
    if os.path.exists(_checkpoint_file):
        with open(_checkpoint_file, 'rb') as f:
            _checkpoint_data = cloudpickle.load(f)
    else:
        _checkpoint_data = dict()


def _dump_checkpoint():
    assert _checkpoint_data is not None, '_checkpoint_data should not be None!'
    with open(_checkpoint_file, 'wb') as f:  # type: ignore
        cloudpickle.dump(_checkpoint_data, f)


def _get_checkpoint(key: str):
    try:
        with _lock:
            _load_checkpoint()
            return _checkpoint_data.get(key, EMPTY)  # type: ignore
    except Exception as e:
        print("Fail to get checkpoint", e)
        return EMPTY


def _set_checkpoint(key: str, value):
    try:
        with _lock:
            _checkpoint_data[key] = value  # type: ignore
            _dump_checkpoint()
    except Exception as e:
        print('Fail to set checkpoint', e)



from typing import List, Callable, TypeVar, Optional
from enum import Enum
from abc import abstractmethod
import time
from .future import IFuture

# Copy from parsl
# TODO:

class JobState(bytes, Enum):
    """Defines a set of states that a job can be in"""
    def __new__(cls, value: int, terminal: bool, status_name: str) -> "JobState":
        obj = bytes.__new__(cls, [value])
        obj._value_ = value
        obj.terminal = terminal
        obj.status_name = status_name
        return obj

    value: int
    terminal: bool
    status_name: str

    UNKNOWN = (0, False, "UNKNOWN")
    PENDING = (1, False, "PENDING")
    RUNNING = (2, False, "RUNNING")
    CANCELLED = (3, True, "CANCELLED")
    COMPLETED = (4, True, "COMPLETED")
    FAILED = (5, True, "FAILED")
    TIMEOUT = (6, True, "TIMEOUT")
    HELD = (7, False, "HELD")

class TimeoutError(RuntimeError):
    ...

class JobFuture(IFuture[JobState]):

    @abstractmethod
    def get_job_state(self) -> JobState:
        ...

    @abstractmethod
    def cancel(self):
        ...

    @abstractmethod
    def redo(self):
        ...

    @abstractmethod
    def get_tries(self) -> int:
        ...

    @abstractmethod
    def is_success(self) -> bool:
        ...

    def is_retriable(self) -> bool:
        return True


DoneFn = Callable[[JobFuture], bool]

def _default_done_fn(_): return True

def retry_fn(max_tries=2) -> DoneFn:

    def callback(job: JobFuture):
        if job.is_success():
            return True
        if job.get_tries() >= max_tries:
            return True
        if job.is_retriable():
            job.redo()
            return False

    return callback  # type: ignore


class GatherJobsFuture(IFuture[List[JobState]]):

    def __init__(self, jobs: List[JobFuture],
                 done_fn: DoneFn = _default_done_fn,
                 raise_exception = True,
                 polling_interval = 10,
                 ):
        self._jobs = jobs
        self._done_fn = done_fn
        self._raise_exception = raise_exception
        self._polling_interval = polling_interval


    def done(self):
        all_done = True
        for job in self._jobs:
            all_done = all_done and job.done() and self._done_fn(job)
        return all_done

    def result(self, timeout: Optional[float] = None):
        if timeout is None:
            timeout = float('inf')
        timeout_ts = time.time() + timeout

        while time.time() < timeout_ts:
            if self.done():
                if self._raise_exception:
                    failed_jobs = [
                        job for job in self._jobs if not job.is_success()]
                    if failed_jobs:
                        raise RuntimeError(
                            'Soem jobs are failed: {}'.format(failed_jobs))
                return [job.result() for job in self._jobs]

            else:
                time.sleep(self._polling_interval)
        else:
            raise TimeoutError('Wait for jobs timeout!')
